"""
Class for reading data created by IGOR Pro
(WaveMetrics, Inc., Portland, OR, USA)

Depends on: igor2 (https://pypi.python.org/pypi/igor2/)

Supported: Read

Author: Andrew Davison
Also contributing: Rick Gerkin

"""

from warnings import warn
import pathlib

import quantities as pq

from neo.io.baseio import BaseIO
from neo.core import Block, Segment, AnalogSignal, NeoReadWriteError


class IgorIO(BaseIO):
    """
    Class for reading Igor Binary Waves (.ibw)
    or Packed Experiment (.pxp) files written by WaveMetrics’
    IGOR Pro software.

    It requires the `igor2` Python package.

    Usage:
        >>> from neo import io
        >>> r = io.IgorIO(filename='...ibw')



    """

    is_readable = True  # This class can only read data
    is_writable = False  # write is not supported
    supported_objects = [Block, Segment, AnalogSignal]
    readable_objects = [Block, Segment, AnalogSignal]
    writeable_objects = []
    has_header = False
    is_streameable = False
    name = "igorpro"
    extensions = ["ibw", "pxp"]
    mode = "file"

    def __init__(self, filename=None, parse_notes=None):
        """


        Arguments:
            filename: the filename

            parse_notes: (optional) A function which will parse the 'notes'
            field in the file header and return a dictionary which will be
            added to the object annotations.

        """
        BaseIO.__init__(self)
        filename = pathlib.Path(filename)
        if filename.suffix[1:] not in self.extensions:
            raise NotImplementedError(f"Only the following extensions are supported: {self.extensions}")
        self.filename = filename
        self.extension = filename.suffix[1:]
        self.parse_notes = parse_notes
        self._filesystem = None

    def read_block(self, lazy=False):
        if lazy:
            raise NeoReadWriteError("This IO does not support lazy reading")

        block = Block(file_origin=str(self.filename))
        block.segments.append(self.read_segment(lazy=lazy))
        return block

    def read_segment(self, lazy=False):
        import igor2.packed as pxp
        from igor2.record.wave import WaveRecord

        if lazy:
            raise NeoReadWriteError("This IO does not support lazy mode")
        segment = Segment(file_origin=str(self.filename))

        if self.extension == "pxp":
            if not self._filesystem:
                _, self.filesystem = pxp.load(str(self.filename))

            def callback(dirpath, key, value):
                if isinstance(value, WaveRecord):
                    signal = self._wave_to_analogsignal(value.wave["wave"], dirpath)
                    segment.analogsignals.append(signal)

            pxp.walk(self.filesystem, callback)
        else:
            segment.analogsignals.append(self.read_analogsignal(lazy=lazy))
        return segment

    def read_analogsignal(self, path=None, lazy=False):
        import igor2.binarywave as bw
        import igor2.packed as pxp

        if lazy:
            raise NeoReadWriteError("This IO does not support lazy mode")

        if self.extension == "ibw":
            data = bw.load(str(self.filename))
            version = data["version"]
            if version > 5:
                raise IOError(f"Igor binary wave file format version {version} " "is not supported.")
        elif self.extension == "pxp":
            if type(path) is not str:
                raise TypeError("A colon-separated Igor-style path must be provided.")
            if not self._filesystem:
                _, self.filesystem = pxp.load(str(self.filename))
                path = path.split(":")
                location = self.filesystem["root"]
                for element in path:
                    if element != "root":
                        location = location[element.encode("utf8")]
            data = location.wave

        return self._wave_to_analogsignal(data["wave"], [])

    def _wave_to_analogsignal(self, content, dirpath):
        if "padding" in content:
            if content["padding"].size != 0:
                raise NeoReadWriteError("Cannot handle non-empty padding")
        signal = content["wData"]
        note = content["note"]
        header = content["wave_header"]
        name = str(header["bname"].decode("utf-8"))
        units = "".join([x.decode() for x in header["dataUnits"]])
        time_units = "".join([x.decode() for x in header["xUnits"]])
        if len(time_units) == 0:
            time_units = "s"
        try:
            t_start = pq.Quantity(header["hsB"], time_units)
        except KeyError:
            t_start = pq.Quantity(header["sfB"][0], time_units)
        try:
            sampling_period = pq.Quantity(header["hsA"], time_units)
        except KeyError:
            sampling_period = pq.Quantity(header["sfA"][0], time_units)
        if self.parse_notes:
            try:
                annotations = self.parse_notes(note)
            except ValueError:
                warn("Couldn't parse notes field.")
                annotations = {"note": note}
        else:
            annotations = {"note": note}
        annotations["igor_path"] = ":".join(item.decode("utf-8") for item in dirpath)

        signal = AnalogSignal(
            signal,
            units=units,
            copy=False,
            t_start=t_start,
            sampling_period=sampling_period,
            name=name,
            file_origin=str(self.filename),
            **annotations,
        )
        return signal


# the following function is to handle the annotations in the
# Igor data files from the Blue Brain Project NMC Portal
def key_value_string_parser(itemsep=";", kvsep=":"):
    """
    Parses a string into a dict.

    Parameters
    ----------
    itemsep : str
        Character which separates items
    kvsep : str
        Character which separates the key and value within an item

    Returns
    -------
    callable
        a function which takes the string to be parsed as the sole argument
        and returns a dict.

    Examples
    --------

        >>> parse = key_value_string_parser(itemsep=";", kvsep=":")
        >>> parse("a:2;b:3")
        {'a': 2, 'b': 3}
    """

    def parser(s):
        items = s.split(itemsep)
        return dict(item.split(kvsep, 1) for item in items if item)

    return parser



"""
Tests of neo.io.wineedrio
"""

import unittest

from neo.io import WinEdrIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestWinedrIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = WinEdrIO
    entities_to_download = ["winedr"]
    entities_to_test = [
        "winedr/File_WinEDR_1.EDR",
        "winedr/File_WinEDR_2.EDR",
        "winedr/File_WinEDR_3.EDR",
    ]


if __name__ == "__main__":
    unittest.main()



#
# neo documentation build configuration file, created by
# sphinx-quickstart on Fri Feb 25 14:18:12 2011.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
import re

from packaging.version import Version

with open("../../pyproject.toml", mode="r") as f:
    txt = f.read()
    neo_release = re.findall('version = "(\S+)"', txt)[0]

neo_version = ".".join((str(e) for e in Version(neo_release).release[:2]))

AUTHORS = "Neo authors and contributors <neuralensemble@googlegroups.com>"

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.append(os.path.abspath('.'))

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "matplotlib.sphinxext.plot_directive",
    "sphinx_gallery.gen_gallery",
    "sphinx_inline_tabs",
    "sphinx_design",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Neo"
copyright = "2010-2024, " + AUTHORS

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = neo_version
# The full version, including alpha/beta/rc tags.
release = neo_release

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = []

# The reST default role (used for this markup: `text`)
# to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme = 'default'
# html_theme = 'sphinxdoc'
# html_theme = 'haiku'
# html_theme = 'scrolls'
# html_theme = 'agogo'
# html_theme = 'alabaster'
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
    "logo": {
        "image_light": "images/neologo_small.png",
        "image_dark": "images/neologo_small_darkmode.png",
    },
    "icon_links": [
        {"name": "GitHub", "url": "https://github.com/NeuralEnsemble/python-neo", "icon": "fa-brands fa-square-github"},
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/neo",
            "icon": "fa-solid fa-box",
        },
    ],
    "icon_links_label": "External Links",  # for screen reader
    "use_edit_page_button": False,
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "images/neologo_small.png"


# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "images/neologo_favicon.png"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["images", "_static"]

html_css_files = [
    "css/custom.css",
]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {'index': 'index.html'}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "neodoc"


# -- Options for LaTeX output -------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
#  documentclass [howto/manual]).
latex_documents = [("index", "neo.tex", "Neo Documentation", AUTHORS, "manual")]

# The name of an image file (relative to this directory) to place at the
# top of the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True

todo_include_todos = True  # set to False before releasing documentation

rst_epilog = """
.. |neo_github_url| replace:: https://github.com/NeuralEnsemble/python-neo/archive/neo-{}.zip
""".format(
    neo_release
)

sphinx_gallery_conf = {
    # 'only_warn_on_example_error': True, # helps with debugging broken examples
    "examples_dirs": "../../examples",  # path to your example scripts
    "gallery_dirs": "examples",  # path to where to save gallery generated output
}



"""

"""

import unittest

import quantities as pq

from neo.io import OpenEphysIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.rawiotest.test_openephysrawio import TestOpenEphysRawIO


class TestOpenEphysIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = OpenEphysIO
    entities_to_download = TestOpenEphysRawIO.entities_to_download
    entities_to_test = TestOpenEphysRawIO.entities_to_test


if __name__ == "__main__":
    unittest.main()



import logging
import os
from pathlib import Path
import unittest

from neo.rawio.neuroscoperawio import NeuroScopeRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO
from neo.test.rawiotest import rawio_compliance as compliance
from neo.utils.datasets import get_local_testing_data_folder


class TestNeuroScopeRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = NeuroScopeRawIO
    entities_to_download = ["neuroscope"]
    entities_to_test = [
        "neuroscope/test1/test1",
        "neuroscope/test1/test1.dat",
        "neuroscope/dataset_1/YutaMouse42-151117.eeg",
    ]

    def test_signal_scale(self):
        local_test_dir = get_local_testing_data_folder()
        fname = os.path.join(local_test_dir, "neuroscope/test1/test1.xml")
        reader = NeuroScopeRawIO(filename=fname)
        reader.parse_header()

        gain = reader.header["signal_channels"][0]["gain"]

        # scale is in mV = range of recording in volts * 1000 mV/V /(number of bits * ampification)
        self.assertAlmostEqual(20.0 * 1000 / (2**16 * 1000), gain)

    def test_binary_argument_with_non_canonical_xml_file(self):

        local_test_dir = get_local_testing_data_folder()
        filename = local_test_dir / "neuroscope/test2/recording.xml"
        binary_file = local_test_dir / "neuroscope/test2/signal1.dat"
        reader = NeuroScopeRawIO(filename=filename, binary_file=binary_file)

        msg = "Before parser_header() no header information should be present"
        assert reader.header is None, msg

        reader.parse_header()

        # After the file resolution test that the right file is being loaded
        assert reader.data_file_path == binary_file
        assert reader.xml_file_path == filename

        txt = reader.__repr__()
        msg = "After parser_header() nb_block should be known"
        assert "nb_block" in txt, msg
        txt = reader._repr_annotations()

        # launch a series of test compliance
        compliance.header_is_total(reader)
        compliance.count_element(reader)
        compliance.read_analogsignals(reader)
        compliance.read_spike_times(reader)
        compliance.read_spike_waveforms(reader)
        compliance.read_events(reader)
        compliance.has_annotations(reader)


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.axographio
"""

from datetime import datetime
import unittest

from neo.io import AxographIO
from neo.test.iotest.common_io_test import BaseTestIO

import numpy as np
from numpy.testing import assert_equal
import quantities as pq
from neo.test.rawiotest.test_axographrawio import TestAxographRawIO


class TestAxographIO(BaseTestIO, unittest.TestCase):
    entities_to_download = ["axograph"]
    entities_to_test = TestAxographRawIO.entities_to_test
    ioclass = AxographIO

    def test_version_1(self):
        """Test reading a version 1 AxoGraph file"""

        filename = self.get_local_path("axograph/AxoGraph_Graph_File")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 1)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Current", "Current"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("pA")
        target = np.array([[-5.5078130], [-3.1171880], [+1.6640626], [+1.6640626], [+4.0546880]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0.0005000000237487257 * pq.s)

        assert_equal(sig.sampling_period, 0.0005000010132789612 * pq.s)

    def test_version_2(self):
        """Test reading a version 2 AxoGraph file"""

        filename = self.get_local_path("axograph/AxoGraph_Digitized_File")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 2)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(
            names,
            [
                "Current",
                "Voltage",
                "Column4",
                "Column5",
                "Column6",
                "Column7",
                "Column8",
                "Column9",
                "Column10",
                "Column11",
                "Column12",
                "Column13",
                "Column14",
                "Column15",
                "Column16",
                "Column17",
                "Column18",
                "Column19",
                "Column20",
                "Column21",
                "Column22",
                "Column23",
                "Column24",
                "Column25",
                "Column26",
                "Column27",
                "Column28",
                "Column29",
            ],
        )

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("pA")
        target = np.array([[0.3125], [9.6875], [9.6875], [9.6875], [9.3750]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0.00009999999747378752 * pq.s)

        assert_equal(sig.sampling_period, 0.00009999999747378750 * pq.s)

    def test_version_5(self):
        """Test reading a version 5 AxoGraph file"""

        filename = self.get_local_path("axograph/AxoGraph_X_File.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 5)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Current", "", "", "", "", ""])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("pA")
        target = np.array([[+3.0846775], [-2.5403225], [-1.2903225], [+6.8346770], [-5.0403230]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0.00005 * pq.s)

        assert_equal(sig.sampling_period, 0.00005 * pq.s)

    def test_version_6(self):
        """Test reading a version 6 AxoGraph file"""

        filename = self.get_local_path("axograph/File_axograph.axgd")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 6)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Membrane Voltage-1"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("mV")
        target = np.array([[-60.731834], [-60.701313], [-60.670795], [-60.701313], [-60.731834]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0.00002 * pq.s)

        assert_equal(sig.sampling_period, 0.00002 * pq.s)

    def test_file_written_by_axographio_package_with_linearsequence(self):
        """Test reading file written by axographio package with linearsequence time column"""

        filename = self.get_local_path("axograph/written-by-axographio-with-linearsequence.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 6)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Data 1", "Data 2"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("mV")
        target = np.array([[0.000000], [9.999833], [19.998667], [29.995500], [39.989334]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0 * pq.s)

        assert_equal(sig.sampling_period, 0.01 * pq.s)

    def test_file_written_by_axographio_package_without_linearsequence(self):
        """Test reading file written by axographio package without linearsequence time column"""

        filename = self.get_local_path("axograph/written-by-axographio-without-linearsequence.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 6)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Data 1", "Data 2"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("mV")
        target = np.array([[0.000000], [9.999833], [19.998667], [29.995500], [39.989334]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0 * pq.s)

        assert_equal(sig.sampling_period, 0.009999999999999787 * pq.s)

    def test_file_with_corrupt_comment(self):
        """Test reading a file with a corrupt comment"""

        filename = self.get_local_path("axograph/corrupt-comment.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.annotations["format_ver"], 6)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["Data 1", "Data 2"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("mV")
        target = np.array([[0.000000], [9.999833], [19.998667], [29.995500], [39.989334]], dtype=np.float32)
        assert_equal(arr, target)

        assert_equal(sig.t_start, 0 * pq.s)

        assert_equal(sig.sampling_period, 0.01 * pq.s)

    def test_multi_segment(self):
        """Test reading an episodic file into multiple Segments"""

        filename = self.get_local_path("axograph/episodic.axgd")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")

        assert_equal(len(blk.segments), 30)
        assert_equal(len(blk.groups), 2)
        assert_equal(len(blk.segments[0].analogsignals), 2)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["CAP", "STIM"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("V")
        target = np.array([[1.37500e-06], [1.53125e-06], [1.34375e-06], [1.09375e-06], [1.21875e-06]], dtype=np.float32)
        assert_equal(arr, target)

    def test_force_single_segment(self):
        """Test reading an episodic file into one Segment"""

        filename = self.get_local_path("axograph/episodic.axgd")
        reader = AxographIO(filename=filename, force_single_segment=True)
        blk = reader.read_block(signal_group_mode="split-all")

        assert_equal(len(blk.segments), 1)
        assert_equal(len(blk.groups), 60)
        assert_equal(len(blk.segments[0].analogsignals), 60)

        names = [sig.name for sig in blk.segments[0].analogsignals]
        assert_equal(names, ["CAP", "STIM"] * 30)

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("V")
        target = np.array([[1.37500e-06], [1.53125e-06], [1.34375e-06], [1.09375e-06], [1.21875e-06]], dtype=np.float32)
        assert_equal(arr, target)

    def test_group_by_same_units(self):
        """Test reading with group-by-same-units"""

        filename = self.get_local_path("axograph/episodic.axgd")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="group-by-same-units")

        assert_equal(len(blk.segments), 30)
        assert_equal(len(blk.groups), 1)
        assert_equal(len(blk.segments[0].analogsignals), 1)

        chan_names = blk.segments[0].analogsignals[0].array_annotations["channel_names"]
        assert_equal(chan_names, ["CAP", "STIM"])

        sig = blk.segments[0].analogsignals[0][:5]
        arr = sig.as_array("V")
        target = np.array(
            [
                [1.37500e-06, 3.43750e-03],
                [1.53125e-06, 2.81250e-03],
                [1.34375e-06, 1.87500e-03],
                [1.09375e-06, 1.56250e-03],
                [1.21875e-06, 1.56250e-03],
            ],
            dtype=np.float32,
        )
        assert_equal(arr, target)

    def test_events_and_epochs(self):
        """Test loading events and epochs"""

        filename = self.get_local_path("axograph/events_and_epochs.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")

        event = blk.segments[0].events[0]
        assert_equal(
            event.times,
            [5999, 5999, 23499, 23499, 26499, 26499, 35999] * blk.segments[0].analogsignals[0].sampling_period,
        )
        assert_equal(event.labels, ["Stop", "Start", "Stop", "Start", "Stop", "Start", "Stop"])

        epoch = blk.segments[0].epochs[0]
        assert_equal(epoch.times, np.array([0.1, 4]) * pq.s)
        assert_equal(epoch.durations, np.array([1.4, 2]) * pq.s)
        assert_equal(epoch.labels, ["test interval 1", "test interval 2"])

    def test_rec_datetime(self):
        """Test parsing the recording datetime from notes"""

        # parsing of rec_datetime differs depending on acquisition mode

        # file obtained in episodic acquisition mode has date and time on
        # separate lines of notes
        filename = self.get_local_path("axograph/episodic.axgd")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.rec_datetime, datetime(2018, 6, 7, 15, 11, 36))

        # file obtained in continuous acquisition mode has date and time in
        # single line of notes
        filename = self.get_local_path("axograph/events_and_epochs.axgx")
        reader = AxographIO(filename=filename)
        blk = reader.read_block(signal_group_mode="split-all")
        assert_equal(blk.rec_datetime, datetime(2019, 5, 25, 20, 16, 25))


if __name__ == "__main__":
    unittest.main()



"""
ImageSequences
==============

"""

##########################################################
# Let's import some packages

from neo.core import ImageSequence
from neo.core import RectangularRegionOfInterest, CircularRegionOfInterest, PolygonRegionOfInterest
import matplotlib.pyplot as plt
import quantities as pq

import random


############################################################
# Now we need to generate some data
# We will just make a nice box and then we can attach this
# ImageSequence to a variety of ROIs
# our ImageSequence will be 50 frames of 100x100 pixel images

l = []
for frame in range(50):
    l.append([])
    for y in range(100):
        l[frame].append([])
        for x in range(100):
            l[frame][y].append(random.randint(0, 50))

#####################################################################
# we then make our image sequence and pull out our results from the
# image_seq

image_seq = ImageSequence(l, sampling_rate=500 * pq.Hz, spatial_scale="m", units="V")

result = image_seq.signal_from_region(
    CircularRegionOfInterest(image_seq, 50, 50, 25),
    CircularRegionOfInterest(image_seq, 10, 10, 5),
    PolygonRegionOfInterest(image_seq, (50, 25), (50, 45), (14, 65), (90, 80)),
)

###############################################################
# It is easy to plot our results using matplotlib

for i in range(len(result)):
    plt.figure()
    plt.plot(result[i].times, result[i])
    plt.xlabel("seconde")
    plt.ylabel("valeur")
plt.tight_layout()
plt.show()



import copy

import numpy as np
import unittest

from neo.core.dataobject import DataObject, _normalize_array_annotations, ArrayDict


class Test_DataObject(unittest.TestCase):
    def test(self):
        pass


class Test_array_annotations(unittest.TestCase):
    def test_check_arr_ann(self):
        # DataObject instance that handles checking
        datobj = DataObject([1, 2])  # Inherits from Quantity, so some data is required

        # Correct annotations
        arr1 = np.asarray(["ABC", "DEF"])
        arr2 = np.asarray([3, 6])
        corr_ann = {"anno1": arr1, "anno2": arr2}

        corr_ann_copy = copy.deepcopy(corr_ann)

        # Checking correct annotations should work fine
        corr_ann = _normalize_array_annotations(corr_ann, datobj._get_arr_ann_length())

        # Make sure the annotations have not been altered
        self.assertSequenceEqual(corr_ann.keys(), corr_ann_copy.keys())
        self.assertTrue((corr_ann["anno1"] == corr_ann_copy["anno1"]).all())
        self.assertTrue((corr_ann["anno2"] == corr_ann_copy["anno2"]).all())

        # Now creating incorrect inputs:

        # Nested dict
        nested_ann = {"anno1": {"val1": arr1}, "anno2": {"val2": arr2}}
        with self.assertRaises(ValueError):
            nested_ann = _normalize_array_annotations(nested_ann, datobj._get_arr_ann_length())

        # Containing None
        none_ann = corr_ann_copy
        # noinspection PyTypeChecker
        none_ann["anno2"] = None
        with self.assertRaises(ValueError):
            none_ann = _normalize_array_annotations(none_ann, datobj._get_arr_ann_length())

        # Multi-dimensional arrays in annotations
        multi_dim_ann = copy.deepcopy(corr_ann)
        multi_dim_ann["anno2"] = multi_dim_ann["anno2"].reshape(1, 2)
        with self.assertRaises(ValueError):
            multi_dim_ann = _normalize_array_annotations(multi_dim_ann, datobj._get_arr_ann_length())

        # Wrong length of annotations
        len_ann = corr_ann
        len_ann["anno1"] = np.asarray(["ABC", "DEF", "GHI"])
        with self.assertRaises(ValueError):
            len_ann = _normalize_array_annotations(len_ann, datobj._get_arr_ann_length())

        # Scalar as array annotation raises Error if len(datobj)!=1
        scalar_ann = copy.deepcopy(corr_ann)
        # noinspection PyTypeChecker
        scalar_ann["anno2"] = 3
        with self.assertRaises(ValueError):
            scalar_ann = _normalize_array_annotations(scalar_ann, datobj._get_arr_ann_length())

        # But not if len(datobj) == 1, then it's wrapped into an array
        # noinspection PyTypeChecker
        scalar_ann["anno1"] = "ABC"
        datobj2 = DataObject([1])
        scalar_ann = _normalize_array_annotations(scalar_ann, datobj2._get_arr_ann_length())
        self.assertIsInstance(scalar_ann["anno1"], np.ndarray)
        self.assertIsInstance(scalar_ann["anno2"], np.ndarray)

        # Lists are also made to np.ndarrays
        list_ann = {"anno1": [3, 6], "anno2": ["ABC", "DEF"]}
        list_ann = _normalize_array_annotations(list_ann, datobj._get_arr_ann_length())
        self.assertIsInstance(list_ann["anno1"], np.ndarray)
        self.assertIsInstance(list_ann["anno2"], np.ndarray)

        # Tuple are also made to np.ndarrays
        list_ann = {"anno1": (3, 6), "anno2": ("ABC", "DEF")}
        list_ann = _normalize_array_annotations(list_ann, datobj._get_arr_ann_length())
        self.assertIsInstance(list_ann["anno1"], np.ndarray)
        self.assertIsInstance(list_ann["anno2"], np.ndarray)

    def test_implicit_dict_check(self):
        # DataObject instance that handles checking
        datobj = DataObject([1, 2])  # Inherits from Quantity, so some data is required

        # Correct annotations
        arr1 = np.asarray(["ABC", "DEF"])
        arr2 = np.asarray([3, 6])
        corr_ann = {"anno1": arr1, "anno2": arr2}

        corr_ann_copy = copy.deepcopy(corr_ann)

        # Implicit checks when setting item in dict directly
        # Checking correct annotations should work fine
        datobj.array_annotations["anno1"] = arr1
        datobj.array_annotations.update({"anno2": arr2})

        # Make sure the annotations have not been altered
        self.assertTrue((datobj.array_annotations["anno1"] == corr_ann_copy["anno1"]).all())
        self.assertTrue((datobj.array_annotations["anno2"] == corr_ann_copy["anno2"]).all())

        # Now creating incorrect inputs:

        # Nested dict
        nested_ann = {"anno1": {"val1": arr1}, "anno2": {"val2": arr2}}
        with self.assertRaises(ValueError):
            datobj.array_annotations["anno1"] = {"val1": arr1}

        # Containing None
        none_ann = corr_ann_copy
        # noinspection PyTypeChecker
        none_ann["anno2"] = None
        with self.assertRaises(ValueError):
            datobj.array_annotations["anno1"] = None

        # Multi-dimensional arrays in annotations
        multi_dim_ann = copy.deepcopy(corr_ann)
        multi_dim_ann["anno2"] = multi_dim_ann["anno2"].reshape(1, 2)
        with self.assertRaises(ValueError):
            datobj.array_annotations.update(multi_dim_ann)

        # Wrong length of annotations
        len_ann = corr_ann
        len_ann["anno1"] = np.asarray(["ABC", "DEF", "GHI"])
        with self.assertRaises(ValueError):
            datobj.array_annotations.update(len_ann)

        # Scalar as array annotation raises Error if len(datobj)!=1
        scalar_ann = copy.deepcopy(corr_ann)
        # noinspection PyTypeChecker
        scalar_ann["anno2"] = 3
        with self.assertRaises(ValueError):
            datobj.array_annotations.update(scalar_ann)

        # But not if len(datobj) == 1, then it's wrapped into an array
        # noinspection PyTypeChecker
        scalar_ann["anno1"] = "ABC"
        datobj2 = DataObject([1])
        datobj2.array_annotations.update(scalar_ann)
        self.assertIsInstance(datobj2.array_annotations["anno1"], np.ndarray)
        self.assertIsInstance(datobj2.array_annotations["anno2"], np.ndarray)

        # Lists are also made to np.ndarrays
        list_ann = {"anno1": [3, 6], "anno2": ["ABC", "DEF"]}
        datobj.array_annotations.update(list_ann)
        self.assertIsInstance(datobj.array_annotations["anno1"], np.ndarray)
        self.assertIsInstance(datobj.array_annotations["anno2"], np.ndarray)

    def test_array_annotate(self):
        # Calls _check_array_annotations, so no need to test for these Errors here
        datobj = DataObject([2, 3, 4])
        arr_ann = {"anno1": [3, 4, 5], "anno2": ["ABC", "DEF", "GHI"]}

        # Pass annotations
        datobj.array_annotate(**arr_ann)

        # Make sure they are correct
        self.assertTrue((datobj.array_annotations["anno1"] == np.array([3, 4, 5])).all())
        self.assertTrue((datobj.array_annotations["anno2"] == np.array(["ABC", "DEF", "GHI"])).all())
        self.assertIsInstance(datobj.array_annotations, ArrayDict)

    def test_arr_anns_at_index(self):
        # Get them, test for desired type and size, content
        datobj = DataObject([1, 2, 3, 4])
        arr_ann = {"anno1": [3, 4, 5, 6], "anno2": ["ABC", "DEF", "GHI", "JKL"]}
        datobj.array_annotate(**arr_ann)

        # Integer as index
        ann_int = datobj.array_annotations_at_index(1)
        self.assertEqual(ann_int, {"anno1": 4, "anno2": "DEF"})
        # Negative integer as index
        ann_int_back = datobj.array_annotations_at_index(-2)
        self.assertEqual(ann_int_back, {"anno1": 5, "anno2": "GHI"})

        # Slice as index
        ann_slice = datobj.array_annotations_at_index(slice(1, 3))
        self.assertTrue((ann_slice["anno1"] == np.array([4, 5])).all())
        self.assertTrue((ann_slice["anno2"] == np.array(["DEF", "GHI"])).all())

        # Slice from beginning to end
        ann_slice_all = datobj.array_annotations_at_index(slice(0, None))
        self.assertTrue((ann_slice_all["anno1"] == np.array([3, 4, 5, 6])).all())
        self.assertTrue((ann_slice_all["anno2"] == np.array(["ABC", "DEF", "GHI", "JKL"])).all())

        # Make sure that original object is edited when editing extracted array_annotations
        ann_slice_all["anno1"][2] = 10
        self.assertEqual(datobj.array_annotations_at_index(2)["anno1"], 10)



"""
Tests of neo.io.axonaio
"""

import unittest

from neo.io.axonaio import AxonaIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.rawiotest.test_axonarawio import TestAxonaRawIO


class TestAxonaIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = AxonaIO
    entities_to_download = ["axona"]

    entities_to_test = TestAxonaRawIO.entities_to_test


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.spikegadgetsrawio import SpikeGadgetsRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestSpikeGadgetsRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = SpikeGadgetsRawIO
    entities_to_download = ["spikegadgets"]
    entities_to_test = [
        "spikegadgets/20210225_em8_minirec2_ac.rec",
        "spikegadgets/W122_06_09_2019_1_fromSD.rec",
        "spikegadgets/SpikeGadgets_test_data_2xNpix1.0_20240318_173658.rec",
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.io import SpikeGadgetsIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestSpikeGadgetsIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = SpikeGadgetsIO
    entities_to_download = ["spikegadgets"]
    entities_to_test = ["spikegadgets/20210225_em8_minirec2_ac.rec", "spikegadgets/W122_06_09_2019_1_fromSD.rec"]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.nixrawio import NIXRawIO

# This class subjects to limitations when there are multiple asymmetric blocks


class NixIO(NIXRawIO, BaseFromRaw):

    name = "NIX IO"

    _prefered_signal_group_mode = "group-by-same-units"
    _prefered_units_group_mode = "all-in-one"

    def __init__(self, filename):
        NIXRawIO.__init__(self, filename)
        BaseFromRaw.__init__(self, filename)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.header = None
        self.file.close()



"""
Class for reading spikegadgets files.
Only continuous signals are supported at the moment.

https://spikegadgets.com/spike-products/

Documentation of the format:
https://bitbucket.org/mkarlsso/trodes/wiki/Configuration

Note :
  * this file format have multiple version. news version include the gain for scaling.
     The actual implementation does not contain this feature because we don't have
     files to test this. So now the gain is "hardcoded" to 1. and so units are
      not handled correctly.

The ".rec" file format contains:
  * a first text part with information in an XML structure
  * a second part for the binary buffer

Author: Samuel Garcia
"""

import numpy as np

from xml.etree import ElementTree

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class SpikeGadgetsRawIO(BaseRawIO):
    extensions = ["rec"]
    rawmode = "one-file"

    def __init__(self, filename="", selected_streams=None):
        """
        Class for reading spikegadgets files.
        Only continuous signals are supported at the moment.

        Initialize a SpikeGadgetsRawIO for a single ".rec" file.

        Parameters
        ----------
        filename: str, default: ''
            The *.rec file to be loaded
        selected_streams: str | list | None, default: None
            sublist of streams to load/expose to API, e.g., ['ECU', 'trodes']
            'trodes' is name for ephy channel (ntrodes)
            None will keep all streams

        Notes
        -----
        This file format has multiple versions:
            - Newer versions include the gain for scaling to microvolts [uV].
            - If the scaling is not found in the header, the gain will be "hardcoded" to 1,
              in which case the units are not handled correctly.
        This will not affect functions that do not rely on the data having physical units,
            e.g., _get_analogsignal_chunk, but functions such as rescale_signal_raw_to_float
            will be inaccurate.

        Examples
        --------
        >>> import neo.rawio
        >>> reader = neo.rawio.SpikeGadgetRawIO(filename='data.rec') # all streams
        # just the electrode channels
        >>> reader_trodes = neo.rawio.SpikeGadgetRawIO(filename='data.rec', selected_streams='trodes')


        """
        BaseRawIO.__init__(self)
        self.filename = filename
        self.selected_streams = selected_streams

    def _source_name(self):
        return self.filename

    def _produce_ephys_channel_ids(self, n_total_channels, n_channels_per_chip):
        """Compute the channel ID labels
        The ephys channels in the .rec file are stored in the following order:
        hwChan ID of channel 0 of first chip, hwChan ID of channel 0 of second chip, ..., hwChan ID of channel 0 of Nth chip,
        hwChan ID of channel 1 of first chip, hwChan ID of channel 1 of second chip, ..., hwChan ID of channel 1 of Nth chip,
        ...
        So if there are 32 channels per chip and 128 channels (4 chips), then the channel IDs are:
        0, 32, 64, 96, 1, 33, 65, 97, ..., 128
        See also: https://github.com/NeuralEnsemble/python-neo/issues/1215
        """
        ephys_channel_ids_list = []
        for hw_channel in range(n_channels_per_chip):
            hw_channel_list = [
                hw_channel + chip * n_channels_per_chip for chip in range(int(n_total_channels / n_channels_per_chip))
            ]
            ephys_channel_ids_list.append(hw_channel_list)
        return [channel for channel_list in ephys_channel_ids_list for channel in channel_list]

    def _parse_header(self):
        # parse file until "</Configuration>"
        header_size = None
        with open(self.filename, mode="rb") as f:
            while True:
                line = f.readline()
                if b"</Configuration>" in line:
                    header_size = f.tell()
                    break

            if header_size is None:
                ValueError("SpikeGadgets: the xml header does not contain '</Configuration>'")

            f.seek(0)
            header_txt = f.read(header_size).decode("utf8")

        # explore xml header
        root = ElementTree.fromstring(header_txt)
        gconf = sr = root.find("GlobalConfiguration")
        hconf = root.find("HardwareConfiguration")
        sconf = root.find("SpikeConfiguration")

        self._sampling_rate = float(hconf.attrib["samplingRate"])
        num_ephy_channels = int(hconf.attrib["numChannels"])

        # check for agreement with number of channels in xml
        sconf_channels = np.sum([len(x) for x in sconf])
        if sconf_channels < num_ephy_channels:
            num_ephy_channels = sconf_channels
        if sconf_channels > num_ephy_channels:
            raise NeoReadWriteError(
                "SpikeGadgets: the number of channels in the spike configuration is larger than the number of channels in the hardware configuration"
            )

        try:
            num_chan_per_chip = int(sconf.attrib["chanPerChip"])
        except KeyError:
            num_chan_per_chip = 32  # default value for Intan chips

        # explore sub stream and count packet size
        # first bytes is 0x55
        packet_size = 1
        stream_bytes = {}
        for device in hconf:
            stream_id = device.attrib["name"]
            if "numBytes" in device.attrib.keys():
                num_bytes = int(device.attrib["numBytes"])
                stream_bytes[stream_id] = packet_size
                packet_size += num_bytes

        # timestamps 4 uint32
        self._timestamp_byte = packet_size
        packet_size += 4

        packet_size += 2 * num_ephy_channels

        # read the binary part lazily
        raw_memmap = np.memmap(self.filename, mode="r", offset=header_size, dtype="<u1")

        num_packet = raw_memmap.size // packet_size
        raw_memmap = raw_memmap[: num_packet * packet_size]
        self._raw_memmap = raw_memmap.reshape(-1, packet_size)

        # create signal channels
        stream_ids = []
        signal_streams = []
        signal_channels = []

        # walk in xml device and keep only "analog" one
        self._mask_channels_bytes = {}
        for device in hconf:
            stream_id = device.attrib["name"]
            for channel in device:

                if "interleavedDataIDByte" in channel.attrib:
                    # TODO LATER: deal with "headstageSensor" which have interleaved
                    continue

                if ("dataType" in channel.attrib.keys()) and (channel.attrib["dataType"] == "analog"):

                    if stream_id not in stream_ids:
                        stream_ids.append(stream_id)
                        stream_name = stream_id
                        signal_streams.append((stream_name, stream_id))
                        self._mask_channels_bytes[stream_id] = []

                    name = channel.attrib["id"]
                    chan_id = channel.attrib["id"]
                    dtype = "int16"
                    # TODO LATER : handle gain correctly according the file version
                    units = ""
                    gain = 1.0
                    offset = 0.0
                    signal_channels.append(
                        (name, chan_id, self._sampling_rate, "int16", units, gain, offset, stream_id)
                    )

                    num_bytes = stream_bytes[stream_id] + int(channel.attrib["startByte"])
                    chan_mask = np.zeros(packet_size, dtype="bool")
                    chan_mask[num_bytes] = True
                    chan_mask[num_bytes + 1] = True
                    self._mask_channels_bytes[stream_id].append(chan_mask)

        if num_ephy_channels > 0:
            stream_id = "trodes"
            stream_name = stream_id
            signal_streams.append((stream_name, stream_id))
            self._mask_channels_bytes[stream_id] = []

            channel_ids = self._produce_ephys_channel_ids(num_ephy_channels, num_chan_per_chip)
            chan_ind = 0
            self.is_scaleable = "spikeScalingToUv" in sconf[0].attrib
            if not self.is_scaleable:
                self.logger.warning(
                    "Unable to read channel gain scaling (to uV) from .rec header. Data has no physical units!"
                )

            for trode in sconf:
                if "spikeScalingToUv" in trode.attrib:
                    gain = float(trode.attrib["spikeScalingToUv"])
                    units = "uV"
                else:
                    gain = 1  # revert to hardcoded gain
                    units = ""

                for schan in trode:
                    chan_id = str(channel_ids[chan_ind])
                    name = "hwChan" + chan_id

                    offset = 0.0
                    signal_channels.append(
                        (name, chan_id, self._sampling_rate, "int16", units, gain, offset, stream_id)
                    )

                    chan_mask = np.zeros(packet_size, dtype="bool")
                    num_bytes = packet_size - 2 * num_ephy_channels + 2 * chan_ind
                    chan_mask[num_bytes] = True
                    chan_mask[num_bytes + 1] = True
                    self._mask_channels_bytes[stream_id].append(chan_mask)

                    chan_ind += 1

        # make mask as array (used in _get_analogsignal_chunk(...))
        self._mask_streams = {}
        for stream_id, l in self._mask_channels_bytes.items():
            mask = np.array(l)
            self._mask_channels_bytes[stream_id] = mask
            self._mask_streams[stream_id] = np.any(mask, axis=0)

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # remove some stream if no wanted
        if self.selected_streams is not None:
            if isinstance(self.selected_streams, str):
                self.selected_streams = [self.selected_streams]
            if not isinstance(self.selected_streams, list):
                raise TypeError(
                    f"`selected_streams` must be of type str or list not of type {type(self.selected_streams)}"
                )

            keep = np.isin(signal_streams["id"], self.selected_streams)
            signal_streams = signal_streams[keep]

            keep = np.isin(signal_channels["stream_id"], self.selected_streams)
            signal_channels = signal_channels[keep]

        # No events channels
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes  channels
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()
        # info from GlobalConfiguration in xml are copied to block and seg annotations
        bl_ann = self.raw_annotations["blocks"][0]
        seg_ann = self.raw_annotations["blocks"][0]["segments"][0]
        for ann in (bl_ann, seg_ann):
            ann.update(gconf.attrib)

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        size = self._raw_memmap.shape[0]
        t_stop = size / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        size = self._raw_memmap.shape[0]
        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]

        raw_unit8 = self._raw_memmap[i_start:i_stop]

        num_chan = len(self._mask_channels_bytes[stream_id])
        re_order = None
        if channel_indexes is None:
            # no loop : entire stream mask
            stream_mask = self._mask_streams[stream_id]
        else:
            # accumulate mask
            if isinstance(channel_indexes, slice):
                chan_inds = np.arange(num_chan)[channel_indexes]
            else:
                chan_inds = channel_indexes

                if np.any(np.diff(channel_indexes) < 0):
                    # handle channel are not ordered
                    sorted_channel_indexes = np.sort(channel_indexes)
                    re_order = np.array([list(sorted_channel_indexes).index(ch) for ch in channel_indexes])

            stream_mask = np.zeros(raw_unit8.shape[1], dtype="bool")
            for chan_ind in chan_inds:
                chan_mask = self._mask_channels_bytes[stream_id][chan_ind]
                stream_mask |= chan_mask

        # this copies the data from the memmap into memory
        raw_unit8_mask = raw_unit8[:, stream_mask]
        shape = raw_unit8_mask.shape
        shape = (shape[0], shape[1] // 2)
        # reshape the and retype by view
        raw_unit16 = raw_unit8_mask.flatten().view("int16").reshape(shape)

        if re_order is not None:
            raw_unit16 = raw_unit16[:, re_order]

        return raw_unit16



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.elanrawio import ElanRawIO


class ElanIO(ElanRawIO, BaseFromRaw):
    """
    Class for reading data from Elan.

    Elan is software for studying time-frequency maps of EEG data.

    Elan is developed in Lyon, France, at INSERM U821

    https://elan.lyon.inserm.fr

    Args:
        filename (string) :
            Full path to the .eeg file
        entfile (string) :
            Full path to the .ent file (optional). If None, the path to the ent
            file is inferred from the filename by adding the ".ent" extension
            to it
        posfile (string) :
            Full path to the .pos file (optional). If None, the path to the pos
            file is inferred from the filename by adding the ".pos" extension
            to it
    """

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename, entfile=None, posfile=None):
        ElanRawIO.__init__(self, filename=filename, entfile=entfile, posfile=posfile)
        BaseFromRaw.__init__(self, filename)



from neo.core.analogsignal import AnalogSignal
import unittest
from neo.core import ImageSequence
from neo.core.regionofinterest import (
    CircularRegionOfInterest,
    RectangularRegionOfInterest,
    PolygonRegionOfInterest,
)
import quantities as pq
import numpy as np
from neo.core import Block, Segment


class TestImageSequence(unittest.TestCase):
    def setUp(self):
        self.data = []
        for frame in range(20):
            self.data.append([])
            for y in range(50):
                self.data[frame].append([])
                for x in range(50):
                    self.data[frame][y].append(x)

    def test_sampling_rate(self):
        # test if error is raise when not giving sample_rate
        with self.assertRaises(ValueError):
            ImageSequence(self.data, units="V", spatial_scale=1 * pq.um)

        with self.assertRaises(TypeError):
            ImageSequence(self.data, frame_duration=500, units="V", spatial_scale=1 * pq.um)
        # test if error is raise when not giving frequency at sampling rate
        with self.assertRaises(TypeError):
            ImageSequence(self.data, units="V", sampling_rate=500, spatial_scale=1 * pq.um)

    def test_error_spatial_scale(self):
        # test if error is raise when not giving spatial scale
        with self.assertRaises(ValueError):
            ImageSequence(self.data, units="V", sampling_rate=500 * pq.Hz)

    def test_units(self):
        with self.assertRaises(TypeError):
            ImageSequence(self.data, units=None, sampling_rate=500 * pq.Hz, spatial_scale=1 * pq.um)

    def test_wrong_dimensions(self):
        seq = ImageSequence(self.data, sampling_rate=500 * pq.Hz, units="V", spatial_scale=1 * pq.um)

        self.assertEqual(seq.sampling_rate, 500 * pq.Hz)
        self.assertEqual(seq.spatial_scale, 1 * pq.um)
        # giving wrong dimension test if it give an error
        with self.assertRaises(ValueError):
            ImageSequence(
                [[0, 1, 2, 4, 2], [0, 1, 2, 4, 5]],
                sampling_rate=500 * pq.Hz,
                units="V",
                spatial_scale=1 * pq.um,
            )

    def test_t_start(self):
        seq = ImageSequence(
            self.data,
            sampling_rate=500 * pq.Hz,
            units="V",
            t_start=250 * pq.ms,
            spatial_scale=1 * pq.um,
        )
        self.assertEqual(seq.t_start, 250 * pq.ms)
        n_frames = seq.shape[0]
        self.assertEqual(seq.duration, n_frames / seq.sampling_rate)
        self.assertEqual(seq.t_stop, 290 * pq.ms)


class TestMethodImageSequence(unittest.TestCase):
    def _create_test_objects(self):
        self.data = []
        for frame in range(25):
            self.data.append([])
            for y in range(5):
                self.data[frame].append([])
                for x in range(5):
                    self.data[frame][y].append(x)
        self.seq = ImageSequence(
            self.data,
            units="V",
            sampling_rate=500 * pq.Hz,
            t_start=250 * pq.ms,
            spatial_scale=1 * pq.um,
        )
        self.rect_ROI = RectangularRegionOfInterest(self.seq, 2, 2, 2, 2)

    def test_signal_from_region(self):
        self._create_test_objects()
        signals = self.seq.signal_from_region(self.rect_ROI)
        self.assertIsInstance(signals, list)
        self.assertEqual(len(signals), 1)
        for signal in signals:
            self.assertIsInstance(signal, AnalogSignal)
            self.assertEqual(signal.t_start, self.seq.t_start)
            self.assertEqual(signal.sampling_period, self.seq.frame_duration)
        with self.assertRaises(ValueError):  # no pixels in region
            zero_size_roi = RectangularRegionOfInterest(self.seq, 1, 1, 0, 0)
            ImageSequence(self.data, units="V", sampling_rate=500 * pq.Hz, spatial_scale=1 * pq.um).signal_from_region(
                zero_size_roi
            )
        with self.assertRaises(ValueError):
            ImageSequence(self.data, units="V", sampling_rate=500 * pq.Hz, spatial_scale=1 * pq.um).signal_from_region()


if __name__ == "__main__":
    unittest.main()



"""
Utility functions to retrieve public datasets.
"""

import os
from pathlib import Path

default_testing_repo = "https://gin.g-node.org/NeuralEnsemble/ephy_testing_data"

global local_testing_data_folder
if os.getenv("EPHY_TESTING_DATA_FOLDER", default=None) is not None:
    local_testing_data_folder = Path(os.getenv("EPHY_TESTING_DATA_FOLDER"))
else:
    # set in home
    local_testing_data_folder = Path.home() / "ephy_testing_data"


def get_local_testing_data_folder():
    global local_testing_data_folder
    return local_testing_data_folder


def download_dataset(repo=default_testing_repo, remote_path=None, local_folder=None):
    """
    Download a dataset with datalad client.

    By default it download the "NeuralEnsemble/ephy_testing_data" on gin platform
    which is used for neo testing.

    Usage:

        download_dataset(
                repo='https://gin.g-node.org/NeuralEnsemble/ephy_testing_data',
                remote_path='blackrock/blackrock_2_1',
                local_folder='/home/myname/Documents/')

    Parameters
    ----------
    repo: str
        The url of the repo.
        If None then 'https://gin.g-node.org/NeuralEnsemble/ephy_testing_data'
        is used
    remote_path: str of Path
        The distant path to retrieve (file or folder)
    local_folder: str or Path or None
        The local folder where to download the data.
        If None, a default project testing folder is used. Default: None

    Returns
    -------
    local_path:
        The local path of the downloaded file or folder
    """
    import datalad.api
    from datalad.support.gitrepo import GitRepo

    if local_folder is None:
        global local_testing_data_folder
        local_folder = local_testing_data_folder
    local_folder = Path(local_folder)

    if local_folder.exists() and GitRepo.is_valid_repo(local_folder):
        dataset = datalad.api.Dataset(path=local_folder)
        # make sure git repo is in clean state
        repo = dataset.repo
        repo.call_git(["checkout", "--force", "master"])
        dataset.update(merge=True)
    else:
        dataset = datalad.api.install(path=local_folder, source=repo)

    if remote_path is None:
        print('Bad boy: you have to provide "remote_path"')
        return

    dataset.get(remote_path)

    local_path = local_folder / remote_path

    return local_path



"""
Class for reading/writing data from micromed (.trc).
Inspired by the Matlab code for EEGLAB from Rami K. Niazy.

Completed with matlab Guillaume BECQ code.

Author: Samuel Garcia
"""

import datetime
import struct
import io

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core import NeoReadWriteError


class StructFile(io.BufferedReader):
    def read_f(self, fmt, offset=None):
        if offset is not None:
            self.seek(offset)
        return struct.unpack(fmt, self.read(struct.calcsize(fmt)))


class MicromedRawIO(BaseRawIO):
    """
    Class for reading  data from micromed (.trc).

    Parameters
    ----------
    filename: str, default: None
        The *.trc file to be loaded
    """

    extensions = ["trc", "TRC"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _parse_header(self):
        with open(self.filename, "rb") as fid:
            f = StructFile(fid)

            # Name
            f.seek(64)
            surname = f.read(22).strip(b" ")
            firstname = f.read(20).strip(b" ")

            # Date
            day, month, year, hour, minute, sec = f.read_f("bbbbbb", offset=128)
            rec_datetime = datetime.datetime(year + 1900, month, day, hour, minute, sec)

            Data_Start_Offset, Num_Chan, Multiplexer, Rate_Min, Bytes = f.read_f("IHHHH", offset=138)

            # header version
            (header_version,) = f.read_f("b", offset=175)
            if header_version != 4:
                raise NotImplementedError(f"`header_version {header_version} is not implemented in neo yet")

            # area
            f.seek(176)
            zone_names = [
                "ORDER",
                "LABCOD",
                "NOTE",
                "FLAGS",
                "TRONCA",
                "IMPED_B",
                "IMPED_E",
                "MONTAGE",
                "COMPRESS",
                "AVERAGE",
                "HISTORY",
                "DVIDEO",
                "EVENT A",
                "EVENT B",
                "TRIGGER",
            ]
            zones = {}
            for zname in zone_names:
                zname2, pos, length = f.read_f("8sII")
                zones[zname] = zname2, pos, length
                if zname != zname2.decode("ascii").strip(" "):
                    raise NeoReadWriteError("expected the zone name to match")

            # raw signals memmap
            sig_dtype = "u" + str(Bytes)
            self._raw_signals = np.memmap(self.filename, dtype=sig_dtype, mode="r", offset=Data_Start_Offset).reshape(
                -1, Num_Chan
            )

            # Reading Code Info
            zname2, pos, length = zones["ORDER"]
            f.seek(pos)
            code = np.frombuffer(f.read(Num_Chan * 2), dtype="u2")

            units_code = {-1: "nV", 0: "uV", 1: "mV", 2: 1, 100: "percent", 101: "dimensionless", 102: "dimensionless"}
            signal_channels = []
            sig_grounds = []
            for c in range(Num_Chan):
                zname2, pos, length = zones["LABCOD"]
                f.seek(pos + code[c] * 128 + 2, 0)

                chan_name = f.read(6).strip(b"\x00").decode("ascii")
                ground = f.read(6).strip(b"\x00").decode("ascii")
                sig_grounds.append(ground)
                logical_min, logical_max, logical_ground, physical_min, physical_max = f.read_f("iiiii")
                (k,) = f.read_f("h")
                units = units_code.get(k, "uV")

                factor = float(physical_max - physical_min) / float(logical_max - logical_min + 1)
                gain = factor
                offset = -logical_ground * factor

                f.seek(8, 1)
                (sampling_rate,) = f.read_f("H")
                sampling_rate *= Rate_Min
                chan_id = str(c)
                stream_id = "0"
                signal_channels.append((chan_name, chan_id, sampling_rate, sig_dtype, units, gain, offset, stream_id))

            signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

            signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

            if np.unique(signal_channels["sampling_rate"]).size != 1:
                raise NeoReadWriteError("The sampling rates must be the same across signal channels")
            self._sampling_rate = float(np.unique(signal_channels["sampling_rate"])[0])

            # Event channels
            event_channels = []
            event_channels.append(("Trigger", "", "event"))
            event_channels.append(("Note", "", "event"))
            event_channels.append(("Event A", "", "epoch"))
            event_channels.append(("Event B", "", "epoch"))
            event_channels = np.array(event_channels, dtype=_event_channel_dtype)

            # Read trigger and notes
            self._raw_events = []
            ev_dtypes = [
                ("TRIGGER", [("start", "u4"), ("label", "u2")]),
                ("NOTE", [("start", "u4"), ("label", "S40")]),
                ("EVENT A", [("label", "u4"), ("start", "u4"), ("stop", "u4")]),
                ("EVENT B", [("label", "u4"), ("start", "u4"), ("stop", "u4")]),
            ]
            for zname, ev_dtype in ev_dtypes:
                zname2, pos, length = zones[zname]
                dtype = np.dtype(ev_dtype)
                rawevent = np.memmap(self.filename, dtype=dtype, mode="r", offset=pos, shape=length // dtype.itemsize)

                keep = (
                    (rawevent["start"] >= rawevent["start"][0])
                    & (rawevent["start"] < self._raw_signals.shape[0])
                    & (rawevent["start"] != 0)
                )
                rawevent = rawevent[keep]
                self._raw_events.append(rawevent)

            # No spikes
            spike_channels = []
            spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

            # fille into header dict
            self.header = {}
            self.header["nb_block"] = 1
            self.header["nb_segment"] = [1]
            self.header["signal_streams"] = signal_streams
            self.header["signal_channels"] = signal_channels
            self.header["spike_channels"] = spike_channels
            self.header["event_channels"] = event_channels

            # insert some annotation at some place
            self._generate_minimal_annotations()
            bl_annotations = self.raw_annotations["blocks"][0]
            seg_annotations = bl_annotations["segments"][0]

            for d in (bl_annotations, seg_annotations):
                d["rec_datetime"] = rec_datetime
                d["firstname"] = firstname
                d["surname"] = surname
                d["header_version"] = header_version

            sig_annotations = self.raw_annotations["blocks"][0]["segments"][0]["signals"][0]
            sig_annotations["__array_annotations__"]["ground"] = np.array(sig_grounds)

    def _source_name(self):
        return self.filename

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals.shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._raw_signals.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(channel_indexes)
        raw_signals = self._raw_signals[slice(i_start, i_stop), channel_indexes]
        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):
        return 0

    def _event_count(self, block_index, seg_index, event_channel_index):
        n = self._raw_events[event_channel_index].size
        return n

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):

        raw_event = self._raw_events[event_channel_index]

        if t_start is not None:
            keep = raw_event["start"] >= int(t_start * self._sampling_rate)
            raw_event = raw_event[keep]

        if t_stop is not None:
            keep = raw_event["start"] <= int(t_stop * self._sampling_rate)
            raw_event = raw_event[keep]

        timestamp = raw_event["start"]
        if event_channel_index < 2:
            durations = None
        else:
            durations = raw_event["stop"] - raw_event["start"]

        try:
            labels = raw_event["label"].astype("U")
        except UnicodeDecodeError:
            # sometimes the conversion do not work : here a simple fix
            labels = np.array([e.decode("cp1252") for e in raw_event["label"]], dtype="U")

        return timestamp, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype) / self._sampling_rate
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        durations = raw_duration.astype(dtype) / self._sampling_rate
        return durations



"""
Tests of neo.io.brainwaresrcio
"""

import logging
import os.path

import unittest

import numpy as np
import quantities as pq

from neo.core import Block, Event, Group, Segment, SpikeTrain
from neo.io import BrainwareSrcIO, brainwaresrcio
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.tools import assert_same_sub_schema, assert_neo_object_is_compliant
from neo.test.iotest.tools import create_generic_reader

FILES_TO_TEST = [
    "brainwaresrc/block_300ms_4rep_1clust_part_ch1.src",
    "brainwaresrcblock_500ms_5rep_empty_fullclust_ch1.src",
    "brainwaresrcblock_500ms_5rep_empty_partclust_ch1.src",
    "brainwaresrcinterleaved_500ms_5rep_ch2.src",
    "brainwaresrcinterleaved_500ms_5rep_nospikes_ch1.src",
    "brainwaresrcinterleaved_500ms_7rep_noclust_ch1.src",
    "brainwaresrclong_170s_1rep_1clust_ch2.src",
    "brainwaresrcmulti_500ms_mulitrep_ch1.src",
    "brainwaresrcrandom_500ms_12rep_noclust_part_ch2.src",
    "brainwaresrcsequence_500ms_5rep_ch2.src",
]

FILES_TO_COMPARE = [
    "brainwaresrc/block_300ms_4rep_1clust_part_ch1",
    "brainwaresrc/block_500ms_5rep_empty_fullclust_ch1",
    "brainwaresrc/block_500ms_5rep_empty_partclust_ch1",
    "brainwaresrc/interleaved_500ms_5rep_ch2",
    "brainwaresrc/interleaved_500ms_5rep_nospikes_ch1",
    "brainwaresrc/interleaved_500ms_7rep_noclust_ch1",
    "",
    "brainwaresrc/multi_500ms_mulitrep_ch1",
    "brainwaresrc/random_500ms_12rep_noclust_part_ch2",
    "brainwaresrc/sequence_500ms_5rep_ch2",
]


def proc_src(filename):
    """Load an src file that has already been processed by the official matlab
    file converter.  That matlab data is saved to an m-file, which is then
    converted to a numpy '.npz' file.  This numpy file is the file actually
    loaded.  This function converts it to a neo block and returns the block.
    This block can be compared to the block produced by BrainwareSrcIO to
    make sure BrainwareSrcIO is working properly

    block = proc_src(filename)

    filename: The file name of the numpy file to load.  It should end with
    '*_src_py?.npz'. This will be converted to a neo 'file_origin' property
    with the value '*.src', so the filename to compare should fit that pattern.
    'py?' should be 'py2' for the python 2 version of the numpy file or 'py3'
    for the python 3 version of the numpy file.

    example: filename = 'file1_src_py2.npz'
             src file name = 'file1.src'
    """
    with np.load(filename, allow_pickle=True) as srcobj:
        srcfile = list(srcobj.items())[0][1]

    filename = os.path.basename(filename[:-12] + ".src")

    block = Block(file_origin=filename)

    NChannels = srcfile["NChannels"][0, 0][0, 0]
    side = str(srcfile["side"][0, 0][0])
    ADperiod = srcfile["ADperiod"][0, 0][0, 0]

    comm_seg = proc_src_comments(srcfile, filename)
    block.segments.append(comm_seg)

    all_units = proc_src_units(srcfile, filename)
    block.groups.extend(all_units)

    for rep in srcfile["sets"][0, 0].flatten():
        proc_src_condition(rep, filename, ADperiod, side, block)

    block.check_relationships()

    return block


def proc_src_comments(srcfile, filename):
    """Get the comments in an src file that has been#!N
    processed by the official
    matlab function.  See proc_src for details"""
    comm_seg = Segment(name="Comments", file_origin=filename)
    commentarray = srcfile["comments"].flatten()[0]
    senders = [res[0] for res in commentarray["sender"].flatten()]
    texts = [res[0] for res in commentarray["text"].flatten()]
    timeStamps = [res[0, 0] for res in commentarray["timeStamp"].flatten()]

    timeStamps = np.array(timeStamps, dtype=np.float32)
    t_start = timeStamps.min()
    timeStamps = pq.Quantity(timeStamps - t_start, units=pq.d).rescale(pq.s)
    texts = np.array(texts, dtype="U")
    senders = np.array(senders, dtype="S")
    t_start = brainwaresrcio.convert_brainwaresrc_timestamp(t_start.tolist())

    comments = Event(times=timeStamps, labels=texts, senders=senders)
    comm_seg.events = [comments]
    comm_seg.rec_datetime = t_start

    return comm_seg


def proc_src_units(srcfile, filename):
    """Get the units in an src file that has been processed by the official
    matlab function.  See proc_src for details"""
    all_units = []
    un_unit = Group(
        name="UnassignedSpikes", file_origin=filename, elliptic=[], boundaries=[], timestamp=[], max_valid=[]
    )

    all_units.append(un_unit)

    sortInfo = srcfile["sortInfo"][0, 0]
    timeslice = sortInfo["timeslice"][0, 0]
    maxValid = timeslice["maxValid"][0, 0]
    cluster = timeslice["cluster"][0, 0]
    if len(cluster):
        maxValid = maxValid[0, 0]
        elliptic = [res.flatten() for res in cluster["elliptic"].flatten()]
        boundaries = [res.flatten() for res in cluster["boundaries"].flatten()]
        fullclust = zip(elliptic, boundaries)
        for ielliptic, iboundaries in fullclust:
            unit = Group(
                file_origin=filename, boundaries=[iboundaries], elliptic=[ielliptic], timeStamp=[], max_valid=[maxValid]
            )
            all_units.append(unit)
    return all_units


def proc_src_condition(rep, filename, ADperiod, side, block):
    """Get the condition in a src file that has been processed by the official
    matlab function.  See proc_src for details"""

    stim = rep["stim"].flatten()
    params = [str(res[0]) for res in stim["paramName"][0].flatten()]
    values = [res for res in stim["paramVal"][0].flatten()]
    stim = dict(zip(params, values))
    sweepLen = rep["sweepLen"][0, 0]

    if not len(rep):
        return

    unassignedSpikes = rep["unassignedSpikes"].flatten()
    if len(unassignedSpikes):
        damaIndexes = [res[0, 0] for res in unassignedSpikes["damaIndex"]]
        timeStamps = [res[0, 0] for res in unassignedSpikes["timeStamp"]]
        spikeunit = [res.flatten() for res in unassignedSpikes["spikes"]]
        respWin = np.array([], dtype=np.int32)
        trains = proc_src_condition_unit(
            spikeunit, sweepLen, side, ADperiod, respWin, damaIndexes, timeStamps, filename
        )
        block.groups[0].spiketrains.extend(trains)
        atrains = [trains]
    else:
        damaIndexes = []
        timeStamps = []
        atrains = []

    clusters = rep["clusters"].flatten()
    if len(clusters):
        IdStrings = [res[0] for res in clusters["IdString"]]
        sweepLens = [res[0, 0] for res in clusters["sweepLen"]]
        respWins = [res.flatten() for res in clusters["respWin"]]
        spikeunits = []
        for cluster in clusters["sweeps"]:
            if len(cluster):
                spikes = [res.flatten() for res in cluster["spikes"].flatten()]
            else:
                spikes = []
            spikeunits.append(spikes)
    else:
        IdStrings = []
        sweepLens = []
        respWins = []
        spikeunits = []

    for unit, IdString in zip(block.groups[1:], IdStrings):
        unit.name = str(IdString)

    fullunit = zip(spikeunits, block.groups[1:], sweepLens, respWins)
    for spikeunit, unit, sweepLen, respWin in fullunit:
        trains = proc_src_condition_unit(
            spikeunit, sweepLen, side, ADperiod, respWin, damaIndexes, timeStamps, filename
        )
        atrains.append(trains)
        unit.spiketrains.extend(trains)

    atrains = zip(*atrains)
    for trains in atrains:
        segment = Segment(
            file_origin=filename, feature_type=-1, go_by_closest_unit_center=False, include_unit_bounds=False, **stim
        )
        block.segments.append(segment)
        segment.spiketrains = trains


def proc_src_condition_unit(spikeunit, sweepLen, side, ADperiod, respWin, damaIndexes, timeStamps, filename):
    """Get the unit in a condition in a src file that has been processed by
    the official matlab function.  See proc_src for details"""
    if not damaIndexes:
        damaIndexes = [0] * len(spikeunit)
        timeStamps = [0] * len(spikeunit)

    trains = []
    for sweep, damaIndex, timeStamp in zip(spikeunit, damaIndexes, timeStamps):
        timeStamp = brainwaresrcio.convert_brainwaresrc_timestamp(timeStamp)
        train = proc_src_condition_unit_repetition(
            sweep, damaIndex, timeStamp, sweepLen, side, ADperiod, respWin, filename
        )
        trains.append(train)
    return trains


def proc_src_condition_unit_repetition(sweep, damaIndex, timeStamp, sweepLen, side, ADperiod, respWin, filename):
    """Get the repetion for a unit in a condition in a src file that has been
    processed by the official matlab function.  See proc_src for details"""
    damaIndex = damaIndex.astype("int32")
    if len(sweep):
        times = np.array([res[0, 0] for res in sweep["time"]])
        shapes = np.concatenate([res.flatten()[np.newaxis][np.newaxis] for res in sweep["shape"]], axis=0)
        trig2 = np.array([res[0, 0] for res in sweep["trig2"]])
    else:
        times = np.array([])
        shapes = np.array([[[]]])
        trig2 = np.array([])

    times = pq.Quantity(times, units=pq.ms, dtype=np.float32)
    t_start = pq.Quantity(0, units=pq.ms, dtype=np.float32)
    t_stop = pq.Quantity(sweepLen, units=pq.ms, dtype=np.float32)
    trig2 = pq.Quantity(trig2, units=pq.ms, dtype=np.uint8)
    waveforms = pq.Quantity(shapes, dtype=np.int8, units=pq.mV)
    sampling_period = pq.Quantity(ADperiod, units=pq.us)

    train = SpikeTrain(
        times=times,
        t_start=t_start,
        t_stop=t_stop,
        trig2=trig2,
        dtype=np.float32,
        timestamp=timeStamp,
        dama_index=damaIndex,
        side=side,
        copy=True,
        respwin=respWin,
        waveforms=waveforms,
        file_origin=filename,
    )
    train.annotations["side"] = side
    train.sampling_period = sampling_period
    return train


class BrainwareSrcIOTestCase(BaseTestIO, unittest.TestCase):
    """
    Unit test testcase for neo.io.BrainwareSrcIO
    """

    ioclass = BrainwareSrcIO
    read_and_write_is_bijective = False

    entities_to_download = ["brainwaresrc"]

    # These are the files it tries to read and test for compliance
    files_to_test = FILES_TO_TEST

    # these are reference files to compare to
    files_to_compare = FILES_TO_COMPARE

    # add the suffix
    for i, fname in enumerate(files_to_compare):
        if fname:
            files_to_compare[i] += "_src_py3.npz"

    def setUp(self):
        super().setUp()

    def test_reading_same(self):
        for ioobj, path in self.iter_io_objects(return_path=True):
            obj_reader_all = create_generic_reader(ioobj, readall=True)
            obj_reader_base = create_generic_reader(ioobj, target=False)
            obj_reader_next = create_generic_reader(ioobj, target="next_block")
            obj_reader_single = create_generic_reader(ioobj)

            obj_all = obj_reader_all()
            obj_base = obj_reader_base()
            obj_single = obj_reader_single()
            obj_next = [obj_reader_next()]
            while ioobj._isopen:
                obj_next.append(obj_reader_next())

            try:
                assert_same_sub_schema(obj_all, obj_base)
                assert_same_sub_schema(obj_all[0], obj_single)
                assert_same_sub_schema(obj_all, obj_next)
            except BaseException as exc:
                exc.args += ("from " + os.path.basename(path),)
                raise

            self.assertEqual(len(obj_all), len(obj_next))

    def test_against_reference(self):
        for filename, refname in zip(self.files_to_test, self.files_to_compare):
            if not refname:
                continue
            obj = self.read_file(filename=filename, readall=True)[0]
            refobj = proc_src(self.get_local_path(refname))
            try:
                assert_neo_object_is_compliant(obj)
                assert_neo_object_is_compliant(refobj)
                # assert_same_sub_schema(obj, refobj)  # commented out until IO is adapted to use Group
            except BaseException as exc:
                exc.args += ("from " + filename,)
                raise


if __name__ == "__main__":
    logger = logging.getLogger(BrainwareSrcIO.__module__ + "." + BrainwareSrcIO.__name__)
    logger.setLevel(100)
    unittest.main()



import unittest

from neo.rawio.spike2rawio import Spike2RawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestSpike2RawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = Spike2RawIO
    entities_to_download = ["spike2"]
    entities_to_test = [
        "spike2/File_spike2_1.smr",
        "spike2/File_spike2_2.smr",
        "spike2/File_spike2_3.smr",
        "spike2/130322-1LY.smr",  # this is for bug 182
        "spike2/multi_sampling.smr",  # this is for bug 466
        "spike2/Two-mice-bigfile-test000.smr",  # SONv9 file
    ]


if __name__ == "__main__":
    unittest.main()



import unittest
from neo.rawio.nixrawio import NIXRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


testfname = ""


class TestNixRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = NIXRawIO
    entities_to_download = ["nix/nixrawio-1.5.nix"]
    entities_to_test = ["nix/nixrawio-1.5.nix"]


if __name__ == "__main__":
    unittest.main()



"""
This module defines :class:`Event`, an array of events.

:class:`Event` derives from :class:`BaseNeo`, from
:module:`neo.core.baseneo`.
"""

from copy import deepcopy

import numpy as np
import quantities as pq

from neo.core.baseneo import merge_annotations
from neo.core.dataobject import DataObject, ArrayDict
from neo.core.epoch import Epoch


def _new_event(
    cls,
    times=None,
    labels=None,
    units=None,
    name=None,
    file_origin=None,
    description=None,
    array_annotations=None,
    annotations=None,
    segment=None,
):
    """
    A function to map Event.__new__ to function that
    does not do the unit checking. This is needed for pickle to work.
    """
    e = Event(
        times=times,
        labels=labels,
        units=units,
        name=name,
        file_origin=file_origin,
        description=description,
        array_annotations=array_annotations,
        **annotations,
    )
    e.segment = segment
    return e


class Event(DataObject):
    """
    Array of events which are the start times of events along with the labels
    of the events

    Parameters
    ----------
    times: quantity array 1d | list
        The times of the events
    labels: numpy.ndarray 1d dtype='U' | list
        Names or labels for the events
    units: quantity units | None, default: None
        If times are list the units of the times
        If times is a quantity array this is ignored
    name: str | None, default: None
        An optional label for the dataset
    description: str | None, default: None
        An optional text descriptoin of the dataset
    file_orgin: str | None, default: None
        The filesystem path or url of the original data file
    array_annotations: dict | None, default: None
        Dict mapping strings to numpy arrays containing annotations for all data points
    **annotations: dict
        Additional user specified metadata stored in the annotations attribue

    Examples
    --------

    >>> from neo.core import Event
    >>> from quantities import s
    >>> import numpy as np
    >>>
    >>> evt = Event(np.arange(0, 30, 10)*s,
    ...             labels=np.array(['trig0', 'trig1', 'trig2'],
    ...                             dtype='U'))
    >>>
    >>> evt.times
    array([  0.,  10.,  20.]) * s
    >>> evt.labels
    array(['trig0', 'trig1', 'trig2'],
              dtype='<U5')

    """

    _parent_objects = ("Segment",)
    _parent_attrs = ("segment",)
    _quantity_attr = "times"
    _necessary_attrs = (("times", pq.Quantity, 1), ("labels", np.ndarray, 1, np.dtype("U")))

    def __new__(
        cls,
        times=None,
        labels=None,
        units=None,
        name=None,
        description=None,
        file_origin=None,
        array_annotations=None,
        **annotations,
    ):
        if times is None:
            times = np.array([]) * pq.s
        elif isinstance(times, (list, tuple)):
            times = np.array(times)
        if len(times.shape) > 1:
            raise ValueError("Times array has more than 1 dimension")
        if labels is None:
            labels = np.array([], dtype="U")
        else:
            labels = np.array(labels)
            if labels.size != times.size and labels.size:
                raise ValueError("Labels array has different length to times")
        if units is None:
            # No keyword units, so get from `times`
            try:
                units = times.units
                dim = units.dimensionality
            except AttributeError:
                raise ValueError("you must specify units")
        else:
            if hasattr(units, "dimensionality"):
                dim = units.dimensionality
            else:
                dim = pq.quantity.validate_dimensionality(units)
        # check to make sure the units are time
        # this approach is much faster than comparing the
        # reference dimensionality
        if len(dim) != 1 or list(dim.values())[0] != 1 or not isinstance(list(dim.keys())[0], pq.UnitTime):
            ValueError(f"Unit {units} has dimensions {dim.simplified}, not [time].")

        obj = pq.Quantity(times, units=dim).view(cls)
        obj._labels = labels
        obj.segment = None
        return obj

    def __init__(
        self,
        times=None,
        labels=None,
        units=None,
        name=None,
        description=None,
        file_origin=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Initialize a new :class:`Event` instance.
        """
        DataObject.__init__(
            self,
            name=name,
            file_origin=file_origin,
            description=description,
            array_annotations=array_annotations,
            **annotations,
        )

    def __reduce__(self):
        """
        Map the __new__ function onto _new_event, so that pickle
        works
        """
        return _new_event, (
            self.__class__,
            np.array(self),
            self.labels,
            self.units,
            self.name,
            self.file_origin,
            self.description,
            self.array_annotations,
            self.annotations,
            self.segment,
        )

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self._labels = getattr(obj, "labels", None)
        self.annotations = getattr(obj, "annotations", None)
        self.name = getattr(obj, "name", None)
        self.file_origin = getattr(obj, "file_origin", None)
        self.description = getattr(obj, "description", None)
        self.segment = getattr(obj, "segment", None)
        # Add empty array annotations, because they cannot always be copied,
        # but do not overwrite existing ones from slicing etc.
        # This ensures the attribute exists
        if not hasattr(self, "array_annotations"):
            self.array_annotations = ArrayDict(self._get_arr_ann_length())

    def __repr__(self):
        """
        Returns a string representing the :class:`Event`.
        """

        objs = [f"{label}@{str(time)}" for label, time in zip(self.labels, self.times)]
        return "<Event: %s>" % ", ".join(objs)

    def _repr_pretty_(self, pp, cycle):
        labels = ""
        if self._labels is not None:
            labels = " with labels"
        pp.text(
            f"{self.__class__.__name__} containing {self.size} events{labels}; "
            f"time units {self.units.dimensionality.string}; datatype {self.dtype} "
        )
        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

    def rescale(self, units, dtype=None):
        """
        Return a copy of the :class:`Event` converted to the specified units

        Parameters
        ----------
        units: quantity units
            The units to convert the Event to
        dtype: None
            Exists for backward compatiblity within quantities see Notes for more info

        Returns
        -------
        obj: neo.core.Event
            A copy of the event with the specified units

        Notes
        -----
        The `dtype` argument exists only for backward compatibility within quantities, see
        https://github.com/python-quantities/python-quantities/pull/204

        """
        # Use simpler functionality, if nothing will be changed
        dim = pq.quantity.validate_dimensionality(units)
        if self.dimensionality == dim:
            return self.copy()

        # Rescale the object into a new object
        obj = self.duplicate_with_new_data(times=self.view(pq.Quantity).rescale(dim), labels=self.labels, units=units)

        # Expected behavior is deepcopy, so deepcopying array_annotations
        obj.array_annotations = deepcopy(self.array_annotations)
        obj.segment = self.segment
        return obj

    @property
    def times(self):
        return pq.Quantity(self)

    def merge(self, other):
        """
        Merge another :class:`Event` into this one.

        Parameter
        ---------
        other: neo.core.Event
            The `Event` to merge into this one

        Notes
        -----
        * The :class:`Event` objects are concatenated horizontally
        (column-wise), :func:`np.hstack`).

        * If the attributes of the two :class:`Event` are not
        compatible, and Exception is raised.
        """
        othertimes = other.times.rescale(self.times.units)
        times = np.hstack([self.times, othertimes]) * self.times.units
        labels = np.hstack([self.labels, other.labels])
        kwargs = {}
        for name in ("name", "description", "file_origin"):
            attr_self = getattr(self, name)
            attr_other = getattr(other, name)
            if attr_self == attr_other:
                kwargs[name] = attr_self
            else:
                kwargs[name] = f"merge({attr_self}, {attr_other})"

        print("Event: merge annotations")
        merged_annotations = merge_annotations(self.annotations, other.annotations)

        kwargs.update(merged_annotations)

        kwargs["array_annotations"] = self._merge_array_annotations(other)

        evt = Event(times=times, labels=labels, **kwargs)

        return evt

    def _copy_data_complement(self, other):
        """
        Copy the metadata from another :class:`Event`.
        Note: Array annotations can not be copied here because length of data can change
        """
        # Note: Array annotations, including labels, cannot be copied
        # because they are linked to their respective timestamps and length of data can be changed
        # here which would cause inconsistencies
        for attr in ("name", "file_origin", "description", "annotations"):
            setattr(self, attr, deepcopy(getattr(other, attr, None)))

    def __getitem__(self, i):
        obj = super().__getitem__(i)
        if self._labels is not None and self._labels.size > 0:
            obj.labels = self._labels[i]
        else:
            obj.labels = self._labels
        try:
            obj.array_annotate(**deepcopy(self.array_annotations_at_index(i)))
            obj._copy_data_complement(self)
        except AttributeError:  # If Quantity was returned, not Event
            obj.times = obj
        return obj

    def set_labels(self, labels):
        if self.labels is not None and self.labels.size > 0 and len(labels) != self.size:
            raise ValueError(f"Labels array has different length to times " f"({len(labels)} != {self.size})")
        self._labels = np.array(labels)

    def get_labels(self):
        return self._labels

    labels = property(get_labels, set_labels)

    def duplicate_with_new_data(self, times, labels, units=None):
        """
        Create a new :class:`Event` with the same metadata
        but different data
        Note: Array annotations can not be copied here because length of data can change
        """
        if units is None:
            units = self.units
        else:
            units = pq.quantity.validate_dimensionality(units)

        new = self.__class__(times=times, units=units)
        new._copy_data_complement(self)
        new.labels = labels
        # Note: Array annotations cannot be copied here, because length of data can be changed
        return new

    def time_slice(self, t_start, t_stop):
        """
        Creates a new `Event` corresponding to the time slice of the original `Event` between (and including) times
        `t_start` and `t_stop`.

        Parameters
        ----------
        t_start: float | None
            The starting time of the time slice
            If None will use -np.inf for the starting time
        t_stop: float | None
            The stopping time of the time slice
            If None will use np.inf for the stopping time

        Returns
        -------
        new_evt: neo.core.Event
            The new `Event` limited by the time points
        """
        _t_start = t_start
        _t_stop = t_stop
        if t_start is None:
            _t_start = -np.inf
        if t_stop is None:
            _t_stop = np.inf

        indices = (self >= _t_start) & (self <= _t_stop)

        # Time slicing should create a deep copy of the object
        new_evt = deepcopy(self[indices])

        return new_evt

    def time_shift(self, t_shift):
        """
        Shifts an :class:`Event` by an amount of time.

        Parameters
        ----------
        t_shift: Quantity (time)
            Amount of time by which to shift the :class:`Event`.

        Returns
        -------
        epoch: Event
            New instance of an :class:`Event` object starting at t_shift later than the
            original :class:`Event` (the original :class:`Event` is not modified).
        """
        new_evt = self.duplicate_with_new_data(times=self.times + t_shift, labels=self.labels)

        # Here we can safely copy the array annotations since we know that
        # the length of the Event does not change.
        new_evt.array_annotate(**self.array_annotations)

        return new_evt

    def to_epoch(self, pairwise=False, durations=None):
        """
        Returns a new Epoch object based on the times and labels in the Event object.

        This method has three modes of action.

        1. By default, an array of `n` event times will be transformed into
           `n-1` epochs, where the end of one epoch is the beginning of the next.
           This assumes that the events are ordered in time; it is the
           responsibility of the caller to check this is the case.
        2. If `pairwise` is True, then the event times will be taken as pairs
           representing the start and end time of an epoch. The number of
           events must be even, otherwise a ValueError is raised.
        3. If `durations` is given, it should be a scalar Quantity or a
           Quantity array of the same size as the Event.
           Each event time is then taken as the start of an epoch of duration
           given by `durations`.

        `pairwise=True` and `durations` are mutually exclusive. A ValueError
        will be raised if both are given.

        If `durations` is given, epoch labels are set to the corresponding
        labels of the events that indicate the epoch start
        If `durations` is not given, then the event labels A and B bounding
        the epoch are used to set the labels of the epochs in the form 'A-B'.
        """

        if pairwise:
            # Mode 2
            if durations is not None:
                raise ValueError("Inconsistent arguments. " "Cannot give both `pairwise` and `durations`")
            if self.size % 2 != 0:
                raise ValueError("Pairwise conversion of events to epochs" " requires an even number of events")
            times = self.times[::2]
            durations = self.times[1::2] - times
            labels = np.array([f"{a}-{b}" for a, b in zip(self.labels[::2], self.labels[1::2])])
        elif durations is None:
            # Mode 1
            times = self.times[:-1]
            durations = np.diff(self.times)
            labels = np.array([f"{a}-{b}" for a, b in zip(self.labels[:-1], self.labels[1:])])
        else:
            # Mode 3
            times = self.times
            labels = self.labels
        return Epoch(times=times, durations=durations, labels=labels)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.spikegadgetsrawio import SpikeGadgetsRawIO


class SpikeGadgetsIO(SpikeGadgetsRawIO, BaseFromRaw):
    __doc__ = SpikeGadgetsRawIO.__doc__

    def __init__(self, filename):
        SpikeGadgetsRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.openephysbinaryrawio import OpenEphysBinaryRawIO


class OpenEphysBinaryIO(OpenEphysBinaryRawIO, BaseFromRaw):
    _prefered_signal_group_mode = "group-by-same-units"
    mode = "dir"

    def __init__(self, dirname):
        OpenEphysBinaryRawIO.__init__(self, dirname=dirname)
        BaseFromRaw.__init__(self, dirname)



import unittest

from neo.rawio.winedrrawio import WinEdrRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestWinEdrRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = WinEdrRawIO
    entities_to_download = ["winedr"]
    entities_to_test = [
        "winedr/File_WinEDR_1.EDR",
        "winedr/File_WinEDR_2.EDR",
        "winedr/File_WinEDR_3.EDR",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.mearecio
"""

import unittest

from neo.io import MEArecIO
from neo.test.iotest.common_io_test import BaseTestIO


try:
    import MEArec as mr

    HAVE_MEAREC = True
except ImportError:
    HAVE_MEAREC = False


@unittest.skipUnless(HAVE_MEAREC, "requires MEArec package")
class TestMEArecIO(BaseTestIO, unittest.TestCase):
    entities_to_download = ["mearec"]
    entities_to_test = ["mearec/mearec_test_10s.h5"]
    ioclass = MEArecIO


if __name__ == "__main__":
    unittest.main()



"""
BaseFromRaw
======

BaseFromRaw implement a bridge between the new neo.rawio API
and the neo.io legacy that give neo.core object.
The neo.rawio API is more restricted and limited and does not cover tricky
cases with asymetrical tree of neo object.
But if a format is done in neo.rawio the neo.io is done for free
by inheritance of this class.
Furthermore, IOs that inherit this BaseFromRaw also have the ability
to lazy load with proxy objects.


"""

import collections
import warnings
import numpy as np

from neo import logging_handler
from neo.core import AnalogSignal, Block, Epoch, Event, IrregularlySampledSignal, Group, Segment, SpikeTrain
from neo.io.baseio import BaseIO

from neo.io.proxyobjects import (
    AnalogSignalProxy,
    SpikeTrainProxy,
    EventProxy,
    EpochProxy,
    ensure_signal_units,
    check_annotations,
    ensure_second,
    proxyobjectlist,
)


import quantities as pq


class BaseFromRaw(BaseIO):
    """
    This implement generic reader on top of RawIO reader.

    Arguments depend on `mode` (dir or file)

    File case::

        reader = BlackRockIO(filename='FileSpec2.3001.nev')

    Dir case::

        reader = NeuralynxIO(dirname='Cheetah_v5.7.4/original_data')

    Other arguments are IO specific.

    """

    is_readable = True
    is_writable = False

    supported_objects = [Block, Segment, AnalogSignal, SpikeTrain, Group, Event, Epoch]
    readable_objects = [Block, Segment]
    writeable_objects = []

    support_lazy = True

    name = "BaseIO"
    description = ""
    extensions = []

    mode = "file"

    _prefered_signal_group_mode = "group-by-same-units"  # 'split-all'

    def __init__(self, *args, **kargs):
        BaseIO.__init__(self, *args, **kargs)
        self.parse_header()

    def read_block(
        self, block_index=0, lazy=False, create_group_across_segment=None, signal_group_mode=None, load_waveforms=False
    ):
        """
        Reads one block of data from a file

        Parameters
        ----------
        block_index: int, default: 0
            In the case of multiple blocks, the block_index specifies which block to read
        lazy: bool, default: False
            Whether to read the block lazily (True) or load into memory (False)
        create_group_across_segment: bool | dict | None, default: None
            If True :
                * Create a neo.Group to group AnalogSignal segments
                * Create a neo.Group to group SpikeTrain across segments
                * Create a neo.Group to group Event across segments
                * Create a neo.Group to group Epoch across segments
            With a dict the behavior can be controlled more finely
                * for example: create_group_across_segment = { 'AnalogSignal': True, 'SpikeTrain': False, ...}
        signal_group_mode: 'split-all' | 'group-by-same-units' | None, default: None
            This control behavior for grouping channels in AnalogSignal.
                * 'split-all': each channel will be give an AnalogSignal
                * 'group-by-same-units' all channel sharing the same quantity units are grouped in a 2D AnalogSignal
            By default None since the default is dependant on the IO
        load_waveforms: bool, default: False
            Determines whether SpikeTrains.waveforms is created

        Returns
        -------
        bl: neo.core.Block
            The block of data

        """

        if signal_group_mode is None:
            signal_group_mode = self._prefered_signal_group_mode

        l = ["AnalogSignal", "SpikeTrain", "Event", "Epoch"]
        if create_group_across_segment is None:
            # @andrew @ julia @michael ?
            # I think here the default None could give this
            create_group_across_segment = {
                "AnalogSignal": True,  # because mimic the old ChannelIndex for AnalogSignals
                "SpikeTrain": False,  # False by default because can create too many object for simulation
                "Event": False,  # not implemented yet
                "Epoch": False,  # not implemented yet
            }
        elif isinstance(create_group_across_segment, bool):
            # bool to dict
            v = create_group_across_segment
            create_group_across_segment = {k: v for k in l}
        elif isinstance(create_group_across_segment, dict):
            # put False to missing keys
            create_group_across_segment = {k: create_group_across_segment.get(k, False) for k in l}
        else:
            raise ValueError("create_group_across_segment must be bool or dict")

        # annotations
        bl_annotations = dict(self.raw_annotations["blocks"][block_index])
        bl_annotations.pop("segments")
        bl_annotations = check_annotations(bl_annotations)

        bl = Block(**bl_annotations)

        # Group for AnalogSignals coming from signal_streams
        if create_group_across_segment["AnalogSignal"]:
            signal_streams = self.header["signal_streams"]
            sub_streams = self.get_sub_signal_streams(signal_group_mode)
            sub_stream_groups = []
            for sub_stream in sub_streams:
                stream_index, inner_stream_channels, name = sub_stream
                group = Group(name=name, stream_id=signal_streams[stream_index]["id"])
                bl.groups.append(group)
                sub_stream_groups.append(group)

        if create_group_across_segment["SpikeTrain"]:
            spike_channels = self.header["spike_channels"]
            st_groups = []
            for c in range(spike_channels.size):
                group = Group(name=f"SpikeTrain group {c}")
                group.annotate(unit_name=spike_channels[c]["name"])
                group.annotate(unit_id=spike_channels[c]["id"])
                bl.groups.append(group)
                st_groups.append(group)

        if create_group_across_segment["Event"]:
            # @andrew @ julia @michael :
            # Do we need this ? I guess yes
            raise NotImplementedError()

        if create_group_across_segment["Epoch"]:
            # @andrew @ julia @michael :
            # Do we need this ? I guess yes
            raise NotImplementedError()

        # Read all segments
        for seg_index in range(self.segment_count(block_index)):
            seg = self.read_segment(
                block_index=block_index,
                seg_index=seg_index,
                lazy=lazy,
                signal_group_mode=signal_group_mode,
                load_waveforms=load_waveforms,
            )
            bl.segments.append(seg)

        # create link between group (across segment) and data objects
        for seg in bl.segments:
            if create_group_across_segment["AnalogSignal"]:
                for c, anasig in enumerate(seg.analogsignals):
                    sub_stream_groups[c].add(anasig)

            if create_group_across_segment["SpikeTrain"]:
                for c, sptr in enumerate(seg.spiketrains):
                    st_groups[c].add(sptr)

        bl.check_relationships()

        return bl

    def read_segment(
        self,
        block_index=0,
        seg_index=0,
        lazy=False,
        signal_group_mode=None,
        load_waveforms=False,
        time_slice=None,
        strict_slicing=True,
    ):
        """
        Reads one segment of the data file

        Parameters
        ----------
        block_index: int, default: 0
            In the case of a multiblock dataset, this specifies which block to read
        seg_index: int, default: 0
            In the case of a multisegment block, this specifies which segmeent to read
        lazy: bool, default: False
            Whether to lazily load the segment (True) or to load the segment into memory (False)
        signal_group_mode: 'split-all' | 'group-by-same-units' | None, default: None
           This control behavior for grouping channels in AnalogSignal.
            * 'split-all': each channel will be give an AnalogSignal
            * 'group-by-same-units' all channel sharing the same quantity units are grouped in a 2D AnalogSignal
        load_waveforms: bool, default: False
            Determines whether SpikeTrains.waveforms is created
        time_slice: tuple[quantity.Quantities | None] | None, default: None
            Whether to take a time slice of the data
                * None: indicates from beginning of the segment to the end of the segment
                * tuple: (t_start, t_stop) with t_start and t_stop being quantities in seconds
                * tuple: (None, t_stop) indicates the beginning of the segment to t_stop
                * tuple: (t_start, None) indicates from t_start to the end of the segment
        strict_slicing: bool, default: True
            Control if an error is raised or not when t_start or t_stop
            is outside of the real time range of the segment.

        Returns
        -------
        seg: neo.core.Segment
            Returns the desired segment based on the parameters given
        """

        if lazy:
            if time_slice is not None:
                raise ValueError("For lazy=True you must specify a time_slice when LazyObject.load(time_slice=...)")
            if load_waveforms:
                raise ValueError(
                    "For lazy=True you must specify load_waveforms when SpikeTrain.load(load_waveforms=...)"
                )

        if signal_group_mode is None:
            signal_group_mode = self._prefered_signal_group_mode

        # annotations
        seg_annotations = self.raw_annotations["blocks"][block_index]["segments"][seg_index].copy()
        for k in ("signals", "spikes", "events"):
            seg_annotations.pop(k)
        seg_annotations = check_annotations(seg_annotations)

        seg = Segment(index=seg_index, **seg_annotations)

        # AnalogSignal
        signal_streams = self.header["signal_streams"]
        sub_streams = self.get_sub_signal_streams(signal_group_mode)
        for sub_stream in sub_streams:
            stream_index, inner_stream_channels, name = sub_stream
            anasig = AnalogSignalProxy(
                rawio=self,
                stream_index=stream_index,
                inner_stream_channels=inner_stream_channels,
                block_index=block_index,
                seg_index=seg_index,
            )
            anasig.name = name

            if not lazy:
                # ... and get the real AnalogSignal if not lazy
                anasig = anasig.load(time_slice=time_slice, strict_slicing=strict_slicing)

            seg.analogsignals.append(anasig)

        # SpikeTrain and waveforms (optional)
        spike_channels = self.header["spike_channels"]
        for spike_channel_index in range(len(spike_channels)):
            # make a proxy...
            sptr = SpikeTrainProxy(
                rawio=self, spike_channel_index=spike_channel_index, block_index=block_index, seg_index=seg_index
            )

            if not lazy:
                # ... and get the real SpikeTrain if not lazy
                sptr = sptr.load(time_slice=time_slice, strict_slicing=strict_slicing, load_waveforms=load_waveforms)
                # TODO magnitude_mode='rescaled'/'raw'

            seg.spiketrains.append(sptr)

        # Events/Epoch
        event_channels = self.header["event_channels"]
        for chan_ind in range(len(event_channels)):
            if event_channels["type"][chan_ind] == b"event":
                e = EventProxy(rawio=self, event_channel_index=chan_ind, block_index=block_index, seg_index=seg_index)
                if not lazy:
                    e = e.load(time_slice=time_slice, strict_slicing=strict_slicing)
                seg.events.append(e)
            elif event_channels["type"][chan_ind] == b"epoch":
                e = EpochProxy(rawio=self, event_channel_index=chan_ind, block_index=block_index, seg_index=seg_index)
                if not lazy:
                    e = e.load(time_slice=time_slice, strict_slicing=strict_slicing)
                seg.epochs.append(e)

        seg.check_relationships()
        return seg

    def get_sub_signal_streams(self, signal_group_mode="group-by-same-units"):
        """
        When signal streams don't have homogeneous SI units across channels,
        they have to be split in sub streams to construct AnalogSignal objects with unique units.

        For backward compatibility (neo version <= 0.5) sub-streams can also be
        used to generate one AnalogSignal per channel.
        """
        signal_streams = self.header["signal_streams"]
        signal_channels = self.header["signal_channels"]

        sub_streams = []
        for stream_index in range(len(signal_streams)):
            stream_id = signal_streams[stream_index]["id"]
            stream_name = signal_streams[stream_index]["name"]
            mask = signal_channels["stream_id"] == stream_id
            channels = signal_channels[mask]
            if signal_group_mode == "group-by-same-units":
                # this does not keep the original order
                _, idx = np.unique(channels["units"], return_index=True)
                all_units = channels["units"][np.sort(idx)]

                if len(all_units) == 1:
                    # no substream
                    # None iwill be transform as slice later
                    inner_stream_channels = None
                    name = stream_name
                    sub_stream = (stream_index, inner_stream_channels, name)
                    sub_streams.append(sub_stream)
                else:
                    for units in all_units:
                        (inner_stream_channels,) = np.nonzero(channels["units"] == units)
                        chan_names = channels[inner_stream_channels]["name"]
                        name = "Channels: (" + " ".join(chan_names) + ")"
                        sub_stream = (stream_index, inner_stream_channels, name)
                        sub_streams.append(sub_stream)
            elif signal_group_mode == "split-all":
                # mimic all neo <= 0.5 behavior
                for i, channel in enumerate(channels):
                    inner_stream_channels = [i]
                    name = channels[i]["name"]
                    sub_stream = (stream_index, inner_stream_channels, name)
                    sub_streams.append(sub_stream)
            else:
                raise (NotImplementedError)

        return sub_streams
